#!/usr/bin/perl

use strict;
use Test::More;

eval "use Test::Pod::Coverage 1.00";
plan $@ ?
    (skip_all => "Test::Pod::Coverage 1.00 required for testing pod coverage") :
    (tests => 6);

pod_coverage_ok("VCS::CMSynergy");
pod_coverage_ok("VCS::CMSynergy::Client",
    { also_private => [ qr/^ccm$/, qr/^start$/ ] });
pod_coverage_ok("VCS::CMSynergy::Helper");
pod_coverage_ok("VCS::CMSynergy::Object");
pod_coverage_ok("VCS::CMSynergy::Project");
pod_coverage_ok("VCS::CMSynergy::Users");

